/*
 * Copyright 2018 Mike Gabriel <mike.gabriel@das-netzwerkteam.de>
 * Copyright 2021 Robert Tari <robert@tari.in>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __INDICATOR_POWER_UTILS_H__
#define __INDICATOR_POWER_UTILS_H__

#include <glib.h>
#include <glib/gi18n.h>
#include <string.h>

void utils_handle_settings_request(void);

#endif /* __INDICATOR_POWER_UTILS_H__ */
