/*

A simple Device structure used internally by indicator-power

Copyright 2012 Canonical Ltd.
Copyright 2022-2023 Robert Tari

Authors:
    Charles Kerr <charles.kerr@canonical.com>
    Robert Tari <robert@tari.in>

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
version 3.0 as published by the Free Software Foundation.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License version 3.0 for more details.

You should have received a copy of the GNU General Public
License along with this library. If not, see
<http://www.gnu.org/licenses/>.
*/

#ifndef __INDICATOR_POWER_DEVICE_H__
#define __INDICATOR_POWER_DEVICE_H__

#include <gio/gio.h> /* GIcon */

G_BEGIN_DECLS

#define INDICATOR_POWER_DEVICE_TYPE            (indicator_power_device_get_type ())
#define INDICATOR_POWER_DEVICE(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), INDICATOR_POWER_DEVICE_TYPE, IndicatorPowerDevice))
#define INDICATOR_POWER_DEVICE_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass),  INDICATOR_POWER_DEVICE_TYPE, IndicatorPowerDeviceClass))
#define INDICATOR_IS_POWER_DEVICE(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), INDICATOR_POWER_DEVICE_TYPE))
#define INDICATOR_IS_POWER_DEVICE_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass),  INDICATOR_POWER_DEVICE_TYPE))
#define INDICATOR_POWER_DEVICE_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj),  INDICATOR_POWER_DEVICE_TYPE, IndicatorPowerDeviceClass))

typedef struct _IndicatorPowerDevice IndicatorPowerDevice;
typedef struct _IndicatorPowerDeviceClass IndicatorPowerDeviceClass;
typedef struct _IndicatorPowerDevicePrivate IndicatorPowerDevicePrivate;

#define INDICATOR_POWER_DEVICE_KIND         "kind"
#define INDICATOR_POWER_DEVICE_MODEL        "model"
#define INDICATOR_POWER_DEVICE_STATE        "state"
#define INDICATOR_POWER_DEVICE_OBJECT_PATH  "object-path"
#define INDICATOR_POWER_DEVICE_PERCENTAGE   "percentage"
#define INDICATOR_POWER_DEVICE_TIME         "time"
#define INDICATOR_POWER_DEVICE_POWER_SUPPLY "power-supply"

typedef enum
{
  UP_DEVICE_KIND_UNKNOWN,
  UP_DEVICE_KIND_LINE_POWER,
  UP_DEVICE_KIND_BATTERY,
  UP_DEVICE_KIND_UPS,
  UP_DEVICE_KIND_MONITOR,
  UP_DEVICE_KIND_MOUSE,
  UP_DEVICE_KIND_KEYBOARD,
  UP_DEVICE_KIND_PDA,
  UP_DEVICE_KIND_PHONE,
  UP_DEVICE_KIND_MEDIA_PLAYER,
  UP_DEVICE_KIND_TABLET,
  UP_DEVICE_KIND_COMPUTER,
  UP_DEVICE_KIND_GAMING_INPUT,
  UP_DEVICE_KIND_PEN,
  UP_DEVICE_KIND_TOUCHPAD,
  UP_DEVICE_KIND_MODEM,
  UP_DEVICE_KIND_NETWORK,
  UP_DEVICE_KIND_HEADSET,
  UP_DEVICE_KIND_SPEAKERS,
  UP_DEVICE_KIND_HEADPHONES,
  UP_DEVICE_KIND_VIDEO,
  UP_DEVICE_KIND_OTHER_AUDIO,
  UP_DEVICE_KIND_REMOTE_CONTROL,
  UP_DEVICE_KIND_PRINTER,
  UP_DEVICE_KIND_SCANNER,
  UP_DEVICE_KIND_CAMERA,
  UP_DEVICE_KIND_WEARABLE,
  UP_DEVICE_KIND_TOY,
  UP_DEVICE_KIND_BLUETOOTH_GENERIC,
  UP_DEVICE_KIND_LAST
}
UpDeviceKind;

typedef enum
{
  UP_DEVICE_STATE_UNKNOWN,
  UP_DEVICE_STATE_CHARGING,
  UP_DEVICE_STATE_DISCHARGING,
  UP_DEVICE_STATE_EMPTY,
  UP_DEVICE_STATE_FULLY_CHARGED,
  UP_DEVICE_STATE_PENDING_CHARGE,
  UP_DEVICE_STATE_PENDING_DISCHARGE,
  UP_DEVICE_STATE_LAST
}
UpDeviceState;


/**
 * IndicatorPowerDeviceClass:
 * @parent_class: #GObjectClass
 */
struct _IndicatorPowerDeviceClass
{
  GObjectClass parent_class;
};

/**
 * IndicatorPowerDevice:
 * @parent: #GObject
 * @priv: A cached reference to the private data for the instance.
*/
struct _IndicatorPowerDevice
{
  GObject parent;
  IndicatorPowerDevicePrivate * priv;
};

/***
****
***/

GType indicator_power_device_get_type (void);

IndicatorPowerDevice* indicator_power_device_new (const gchar    * object_path,
                                                  UpDeviceKind     kind,
                                                  const gchar    * model,
                                                  gdouble          percentage,
                                                  UpDeviceState    state,
                                                  time_t           time,
                                                  gboolean         power_supply);

/**
 * Convenience wrapper around indicator_power_device_new()
 * @variant holds the same args as indicator_power_device_new() in "(sussdut)"
 */
IndicatorPowerDevice* indicator_power_device_new_from_variant (GVariant * variant);


UpDeviceKind  indicator_power_device_get_kind              (const IndicatorPowerDevice * device);
const gchar * indicator_power_device_get_model             (const IndicatorPowerDevice * device);
UpDeviceState indicator_power_device_get_state             (const IndicatorPowerDevice * device);
const gchar * indicator_power_device_get_object_path       (const IndicatorPowerDevice * device);
gdouble       indicator_power_device_get_percentage        (const IndicatorPowerDevice * device);
time_t        indicator_power_device_get_time              (const IndicatorPowerDevice * device);
gboolean      indicator_power_device_get_power_supply      (const IndicatorPowerDevice * device);

GStrv         indicator_power_device_get_icon_names        (const IndicatorPowerDevice * device, gboolean panel);
GIcon       * indicator_power_device_get_gicon             (const IndicatorPowerDevice * device, gboolean panel, gboolean bShowCharge);


char        * indicator_power_device_get_readable_text     (const IndicatorPowerDevice * device, gboolean bModelName);

char        * indicator_power_device_get_accessible_text   (const IndicatorPowerDevice * device);

char        * indicator_power_device_get_readable_title    (const IndicatorPowerDevice * device,
                                                            gboolean                     want_time,
                                                            gboolean                     want_percent);

char        * indicator_power_device_get_accessible_title  (const IndicatorPowerDevice * device,
                                                            gboolean                     want_time,
                                                            gboolean                     want_percent);


G_END_DECLS

#endif
